; fibon.asm - written for Win32 MASM by Phil at masmforum.com - June 7, 2005
;
; Based on ideas from Ed Beroset's fibo.asm program at www.beroset.com
; With code and suggestions contributed by masmforum.com members Vortex,
; AeroASM, Mirno and many others. Also, Ed Beroset and Frank Kotrel
; contributed suggestions for optimizing the original fibo.asm program
; and I have implemented many of their ideas in the following code.

;****************************************************************************
; This comment and many ideas from Ed Beroset's fibo.asm at www.beroset.com
;****************************************************************************
; Number digits to use.  This is based on a little bit of tricky math.
; One way to calculate F(n) (i.e. the nth term of the Fibonacci seeries)
; is to use the equation int(phi^n/sqrt(5)) where ^ means exponentiation
; and phi = (1 + sqrt(5))/2, the "golden number" which is a constant about 
; equal to 1.618.  To get the number of decimal digits, we just take the 
; base ten log of this number.  We can very easily see how to get the 
; base phi log of F(n) -- it's just n*lp(phi)+lp(sqrt(5)), where lp means 
; a base phi log.  To get the base ten log of this we just divide by the 
; base ten log of phi.  If we work through all that math, we get:
;
; digits = terms * log(phi) + log(sqrt(5))/log(phi)
;
; The formulas below are slightly high to assure that we always have 
; enough room. The 150000th term has 31348 digits, but this formula
; gives 31351. Not too much waste and a little extra.
;
;       digits  equ (maxTerms*209+1673)/1000 
;
; Likewise, determine how many dwords of storage are required to store the
; maximum requested term. Each dword holds 8 packed digits but make sure we
; round up to ensure a little extra storage.
;
;       ddsize  equ (digits+3)/4
;
;****************************************************************************
; assemble and link as a Win32 console application with:
;****************************************************************************
;
; ml   /c /coff fibon.asm
; link /subsystem:console fibon.obj
;
; command line usage: fibon [optional term]
;

.386
.model flat,stdcall
option casemap:none

.nolist

include     \masm32\include\windows.inc
include     \masm32\include\kernel32.inc
include     \masm32\include\user32.inc
include     \masm32\include\masm32.inc
includelib  \masm32\lib\kernel32.lib
includelib  \masm32\lib\user32.lib
includelib  \masm32\lib\masm32.lib

include     \masm32\macros\macros.asm

.list

ifndef  MAX
        maxTerms equ 300
else
        maxTerms equ (MAX)
endif

ifndef  MIN
        minTerms equ 1
else
        minTerms equ (MIN)
endif

;****************************************************************************
; Initialized Data
;****************************************************************************
.data
hStatus dd NO_ERROR                     ; Exit status
header  db 'Fibonacci(0):   '           ; Last characters must be '(0):   '
hdrlen equ $ - header                   ; Must be 16 characters
usemsg  db 'Usage: fibon [term]',13,10
uselen equ $ - usemsg
errfmt  db 'Error Code %i',13,10
        db 'Description: %s',0
badmsg  db 'Unknown Error',13,10
badlen equ $ - badmsg

        align 4
min     dd minTerms-1                   ; First Fibonacci term to display - 1
max     dd maxTerms                     ; Last Fibonacci term to display
;****************************************************************************
; Uninitialized Data
;****************************************************************************
.data?
hStdOut dd ?                            ; standard output file handle
hStdErr dd ?                            ; standard error file handle
bLimit  dd ?                            ; limit on number of bytes to write
bRequest dd ?                           ; number of bytes in write request
bWritten dd ?                           ; number of bytes written
pMem    dd ?                            ; pointer to allocated memory
pNum1   dd ?                            ; pointer to first number
pNum2   dd ?                            ; pointer to second number
pFirst  dd ?                            ; pointer to first free character
pString dd ?                            ; pointer to output string
digits  dd ?                            ; maximum number of digits required
ddsize  dd ?                            ; number of dwords required per term
wksize  dd ?                            ; current working size (# of dwords)

;****************************************************************************
; Initialization and Command Line Argument Parsing
;****************************************************************************
.code
start:  push    ebp
        mov     ebp,esp
        sub     esp,512+80
        lea     esi,[ebp-80]            ; argv char pointer array
        lea     edi,[ebp-592]           ; argument string storage
        invoke  GetStdHandle,STD_OUTPUT_HANDLE
        mov     hStdOut,eax
        mov     ebx,-1                  ; Unlimited writes for disk files
        invoke  GetFileType,hStdOut
        cmp     eax,FILE_TYPE_DISK
        je      @F
        mov     ebx,32768               ; Write request limit for non-disk
@@:     mov     bLimit,ebx
        invoke  GetStdHandle,STD_ERROR_HANDLE
        mov     hStdErr,eax
        invoke  GetCommandLine          ; Adapted from Startc.asm written by
        lea     edx,[eax-1]             ;  Vortex at masmforum.com
        xor     eax,eax
        mov     ch,32
scan:
	inc	edx
	mov	cl,byte PTR [edx]
	or	cl,cl
	jz	finish
	cmp	cl,32
	je	scan

	inc	eax
        mov     dword PTR [esi],edi
	add	esi,4
restart:
	mov	cl,byte PTR [edx]
	or 	cl,cl
	jz	finish
	cmp	cl,ch
	je	end_of_line
	cmp	cl,34
	jne	@f
	xor	ch,32
	jmp	next_char
@@:	
	mov	byte PTR [edi],cl
	inc	edi
next_char:
	inc	edx
	jmp	restart
end_of_line:
	mov	byte PTR [edi],0
	inc	edi
	jmp	scan	
finish:
        mov     byte PTR [edi],0
        lea     ebx,[ebp-80]
        call    main
        mov     esp,ebp
        pop     ebp
        invoke  ExitProcess,eax

;****************************************************************************
; Main Fibonacci Initialization and Argument Processing
;  called from console initialization code above with
;     eax = number of command line parameters
;     ebx = offset to parameter string array
;  return must restore calling ebp
;     eax = exit status (0 = NO_ERROR)
;****************************************************************************

main:   push    ebp
        dec     eax
        jz      noargs
        mov     edx,[ebx+4]             ; Convert 1st parameter to dword
        xor     ebx,ebx                 ;  and set min-max for single term
        xor     ecx,ecx
@@:     mov     cl,byte PTR [edx]
        or      cl,cl
        jz      @F
        sub     cl,'0'
        cmp     cl,9
        ja      usage                   ; Must be a numeric string
        lea     ebx,[ebx+4*ebx]
        lea     ebx,[ecx+2*ebx]
        add     edx,1
        jmp     @B
@@:     mov     max,ebx
        dec     ebx
        mov     min,ebx

;****************************************************************************
; Allocate Global Storage for two numbers and string
;****************************************************************************

noargs: mov     eax,max
        mov     ebx,209
        mul     ebx
        add     eax,1673
        mov     ebx,1000
        div     ebx
        mov     digits,eax              ; digits=(maxTerms*209+1673)/1000 
        add     eax,7
        shr     eax,3
        mov     ddsize,eax              ; ddsize=(digits+3)/4
        shl     eax,3
        add     eax,hdrlen
        add     eax,digits
        add     eax,8                   ; memsize=(8*ddsize+hdrlen+digits+8)
        invoke  GlobalAlloc,GPTR,eax
        or      eax,eax
        jz      error

;****************************************************************************
; Fibonacci Loop Setup
;****************************************************************************

        mov     pMem,eax
        mov     pNum1,eax
        mov     ebx,ddsize
        shl     ebx,2
        add     eax,ebx
        mov     pNum2,eax
        add     eax,ebx
        mov     pString,eax
        mov     esi,OFFSET header       ; copy 16-byte header to string
        mov     edi,pString
        movsd
        movsd
        movsd
        movsd
        sub     edi, 2                  ; first free char is 14th byte
        mov     pFirst,edi
        mov     esi,pNum1
        mov     edi,pNum2
        mov     eax,1
        mov     [esi],eax
        mov     wksize,1                ; Initial working size is 1 dword
        mov     ebp,esp                 ; Create a frame for parameters

;****************************************************************************
; Fibonacci Processing Loop
;****************************************************************************

top:    push    esi                     ; [ebp-4] is source operand
        push    edi                     ; [ebp-8] is destination operand
        call    DoTerm
        pop     esi                     ; exchange source and destination
        pop     edi
        jc      error                   ; give it up for carry
        dec     max
        jnz     top
        jnc     done

;****************************************************************************
; Error Processing and Exit
;****************************************************************************

error:  push    ebp
        mov     ebp,esp
        sub     esp,512
        invoke  GetLastError
        mov     hStatus,eax
        cmp     eax,NO_ERROR
        jne     @F
        mov     edi,OFFSET badmsg
        mov     ebx,badlen
        jmp     wrterr

@@:     mov     ebx,eax
        lea     esi,[ebp-256]
        invoke  FormatMessage,
                FORMAT_MESSAGE_FROM_SYSTEM,
                0,
                ebx,0,
                esi,
                256,0

        lea     edi,[ebp-512]
        invoke  wsprintf,edi,OFFSET errfmt,ebx,esi
        mov     ebx,edi
@@:     mov     cl,[ebx]
        or      cl,cl
        jz      @F
        add     ebx,1
        jmp     @B
@@:     sub     ebx,edi

wrterr: invoke  WriteFile,hStdErr,edi,ebx,ADDR bWritten,NULL
        mov     esp,ebp
        pop     ebp

done:   mov     eax,pMem
        or      eax,eax
        jz      fini
        invoke  GlobalFree,eax
fini:   pop     ebp
        mov     eax,hStatus
        ret

usage:  invoke  WriteFile,hStdErr,OFFSET usemsg,uselen,ADDR bWritten,NULL
        mov     eax,-1
        pop     ebp
        ret

;****************************************************************************
; All Fibonacci functions placed inline here to avoid function call overhead
;****************************************************************************

DoTerm:

; 1) AddPackedBCD       ; Generate next term and store in destination
; 2) IncrementCount     ; Update current term sequence number
; 3) CheckMinTerm       ; Return until minimum term sequence is reached
; 4) BuildString        ; Unpack bcd digits and store as ASCII in string
; 5) PrintString        ; Display the string and return to caller

;****************************************************************************
; BCD addition algorithm courtesy Douglas W. Jones at University Of Iowa
; Details available at http://www.cs.uiowa.edu/~jones/bcd/bcd.html#packed
; This is my implementation of his posted C algorithm.
;****************************************************************************

        align 4

AddPackedBCD:
        mov     eax,wksize
        mov     ecx,eax
        shl     eax,2
        neg     ecx
        add     esi,eax
        add     edi,eax
        nop
nocry:  mov     eax,[esi+4*ecx]
        mov     ebx,[edi+4*ecx]
        mov     edx,eax
        add     eax,66666666h
        jc      foul
        add     eax,ebx
        jc      cry2
nocry1: xor     edx,ebx
        xor     edx,eax
        not     edx
        and     edx,11111110h
        mov     ebx,edx
        shr     edx,2
        shr     ebx,3
        or      edx,ebx
        or      edx,60000000h
        sub     eax,edx
        mov     [edi+4*ecx],eax
        add     ecx,1
        jnz     nocry
        jmp short IncrementCount

        align 4

cry:    mov     eax,[esi+4*ecx]
        mov     ebx,[edi+4*ecx]
        mov     edx,eax
        add     eax,66666667h
        jnc     cry1
        add     eax,ebx
        jmp short cry2
        nop
        nop
cry1:   add     eax,ebx
        jnc     nocry1
cry2:   xor     edx,ebx
        xor     edx,eax
        not     edx
        and     edx,11111110h
        mov     ebx,edx
        shr     edx,2
        shr     ebx,3
        or      edx,ebx
        sub     eax,edx
        mov     [edi+4*ecx],eax
        add     ecx,1
        jnz     cry        
        mov     eax,wksize
        cmp     eax,ddsize
        jae     foul
        add     wksize,1
        mov dword ptr [edi+4*ecx],1
        jmp short IncrementCount

foul:   stc
        ret

;****************************************************************************
; Increment the ASCII header count in place
;****************************************************************************
        align   4

IncrementCount:
        mov     esi,pFirst
        sub     esi,4                   ; esi -> least significant digit
        mov     al,[esi]
i1:     add     al, 1                   ; increment ASCII digit
        cmp     al,'9'
        jbe     iexit
        mov byte ptr [esi],'0'          ; store a zero if we had a carry
        sub     esi,1
        mov     al,[esi]                ; do next more significant digit
        cmp     al,'('
        jne     i1                      ; propogate carry thru all digits
        mov     al,'1'                  ; store a one for final carry
i2:     add     esi,1                   ; extend counter as required
        xchg    [esi],al
        cmp     esi,pFirst
        jne     i2
        add     esi,1
        add     pFirst,1                ; increment first free char pointer
iexit:  mov     [esi],al                ; store updated character and exit

;****************************************************************************
; Return for all terms less than minimum
;****************************************************************************

CheckMinTerm:
        dec     min
        js      BuildString
        clc
        ret

;****************************************************************************
; Unpack BCD and build numeric ASCII term for display
;****************************************************************************
        align   4

BuildString:
        mov     esi,[ebp-4]             ; convert source operand
        mov     edi,pFirst              ; store beginning at first free char
        mov     edx,wksize
skiplz: sub     edx,1                   ; skip leading zeros
        mov     ebx,[esi+4*edx]
        bsr     eax,ebx                 ; find first one bit (nn = 31 to 0)
        jz      skiplz
        mov     ecx,01Ch                ; maximum shift count and mask is 28
        and     eax,ecx
        sub     ecx,eax                 ; shift 28-(nn & 11100) multiple of 4
        jz      bld8
        rol     ebx,cl                  ; rotate significant digits to bit 31
        mov     ecx,eax
        shr     ecx,2
@@:     rol     ebx,4                   ; convert first 1-7 digits
        mov     al,0Fh
        and     al,bl
        add     al,30h
        mov     [edi],al                ; store ASCII digit and bump pointer
        add     edi,1
        sub     ecx,1
        jns     @B
        jmp     bld0                    ; do remaining 8-digits multiples

        align   4

bld8:   mov     eax, ebx                ; quick algorithm sugessted by Mirno
        mov     ecx, ebx
        and     eax, 0F0F0F0Fh
        shr     ecx, 4
        add     eax, 30303030h
        and     ecx, 0F0F0F0Fh
        add     edi, 8
        add     ecx, 30303030h
        mov     [edi - 1], al
        mov     [edi - 2], cl
        mov     [edi - 3], ah
        mov     [edi - 4], ch
        shr     eax, 16
        shr     ecx, 16
        mov     [edi - 5], al
        mov     [edi - 6], cl
        mov     [edi - 7], ah
        mov     [edi - 8], ch
bld0:   sub     edx,1
        mov     ebx,[esi+4*edx]
        jns     bld8

;****************************************************************************
; Finally, add the <crlf> to destination string, calculate length, then print
; and return to main to see if we're done
;****************************************************************************

EndOfLine:
        mov word ptr [edi],0A0Dh
        add     edi,2
        mov     ebx,edi
        mov     esi,pString
        sub     ebx,esi

WriteString:
        cmp     ebx,bLimit
        ja      limit
        invoke  WriteFile,hStdOut,esi,ebx,ADDR bWritten,NULL
        or      eax,eax
        jnz     okay
fail:   stc
okay:   ret

limit:  mov     bRequest,ebx
        mov     ebx,bLimit
        invoke  WriteFile,hStdOut,esi,ebx,ADDR bWritten,NULL
        or      eax,eax
        jz      fail
        mov     eax,bLimit
        mov     ebx,bRequest
        add     esi,eax
        sub     ebx,eax
        cmp     ebx,eax
        jbe     WriteString
        jmp short limit

        end start
